/* PSPP - a program for statistical analysis.
   Copyright (C) 2011, 2012 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifndef ASCII_H
#define ASCII_H 1

struct output_driver;

void ascii_test_write (struct output_driver *,
                       const char *s, int x, int y, bool bold, bool underline);
void ascii_test_set_length (struct output_driver *, int y, int length);
void ascii_test_flush (struct output_driver *);

#endif /* ascii.h */
