/* -*- c -*- */

/* Summary functions for all variables. */
S(CTSF_COUNT,          "COUNT",      CTFT_UECELL, CTF_COUNT,   CTFA_ALL)
S(CTSF_areaPCT_COUNT,  "PCT.COUNT",  CTFT_AREA,   CTF_PERCENT, CTFA_ALL)
S(CTSF_areaPCT_VALIDN, "PCT.VALIDN", CTFT_AREA,   CTF_PERCENT, CTFA_ALL)
S(CTSF_areaPCT_TOTALN, "PCT.TOTALN", CTFT_AREA,   CTF_PERCENT, CTFA_ALL)

/* Scale variables, totals, and subtotals. */
S(CTSF_MAXIMUM,        "MAXIMUM",    CTFT_CELL,   CTF_GENERAL, CTFA_SCALE)
S(CTSF_MEAN,           "MEAN",       CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_MEDIAN,         "MEDIAN",     CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_MINIMUM,        "MINIMUM",    CTFT_CELL,   CTF_GENERAL, CTFA_SCALE)
S(CTSF_MISSING,        "MISSING",    CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_MODE,           "MODE",       CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_PTILE,          "PTILE",      CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_RANGE,          "RANGE",      CTFT_CELL,   CTF_GENERAL, CTFA_SCALE)
S(CTSF_SEMEAN,         "SEMEAN",     CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_STDDEV,         "STDDEV",     CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_SUM,            "SUM",        CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_TOTALN,         "TOTALN",     CTFT_UECELL, CTF_COUNT,   CTFA_SCALE)
S(CTSF_VALIDN,         "VALIDN",     CTFT_UECELL, CTF_COUNT,   CTFA_SCALE)
S(CTSF_VARIANCE,       "VARIANCE",   CTFT_UCELL,  CTF_GENERAL, CTFA_SCALE)
S(CTSF_areaPCT_SUM,    "PCT.SUM",    CTFT_AREA,   CTF_PERCENT, CTFA_SCALE)

/* Debugging and troubleshooting. */
S(CTSF_areaID,         "ID",         CTFT_AREA,   CTF_COUNT,   CTFA_ALL)
