/* PSPP - a program for statistical analysis.
   Copyright (C) 2006, 2009, 2010, 2011, 2013, 2016 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* Utility commands acceptable anywhere. */
DEF_CMD (S_ANY, F_ENHANCED, "CLOSE FILE HANDLE", N_("Close File Handle"), cmd_close_file_handle)
DEF_CMD (S_ANY, 0, "CACHE", N_("Cache"), cmd_cache)
DEF_CMD (S_ANY, 0, "CD", N_("CD"), cmd_cd)
DEF_CMD (S_ANY, 0, "DEFINE", N_("Define"), cmd_define)
DEF_CMD (S_ANY, 0, "DISPLAY MACROS", N_("Display Macros"), cmd_display_macros)
DEF_CMD (S_ANY, 0, "DO REPEAT", N_("Do Repeat"), cmd_do_repeat)
DEF_CMD (S_ANY, 0, "END REPEAT", N_("End Repeat"), cmd_end_repeat)
DEF_CMD (S_ANY, 0, "ECHO", N_("Echo"), cmd_echo)
DEF_CMD (S_ANY, 0, "ERASE", N_("Erase"), cmd_erase)
DEF_CMD (S_ANY, 0, "EXIT", N_("Exit"), cmd_finish)
DEF_CMD (S_ANY, 0, "FILE HANDLE", N_("File Handle"), cmd_file_handle)
DEF_CMD (S_ANY, 0, "FILE LABEL", N_("File Label"), cmd_file_label)
DEF_CMD (S_ANY, 0, "FINISH", N_("Finish"), cmd_finish)
DEF_CMD (S_ANY, 0, "HOST", N_("Host"), cmd_host)
DEF_CMD (S_ANY, 0, "INCLUDE", N_("Include"), cmd_include)
DEF_CMD (S_ANY, 0, "INSERT", N_("Insert"), cmd_insert)
DEF_CMD (S_ANY, 0, "MATRIX", N_("Matrix"), cmd_matrix)
DEF_CMD (S_ANY, 0, "MCONVERT", N_("Matri Convert"), cmd_mconvert)
DEF_CMD (S_ANY, 0, "N OF CASES", N_("N of Cases"), cmd_n_of_cases)
DEF_CMD (S_ANY, F_ABBREV, "N", N_("N"), cmd_n_of_cases)
DEF_CMD (S_ANY, 0, "NEW FILE", N_("New File"), cmd_new_file)
DEF_CMD (S_ANY, 0, "OUTPUT MODIFY", N_("Output Modify"), cmd_output_modify)
DEF_CMD (S_ANY, 0, "PERMISSIONS", N_("Permissions"), cmd_permissions)
DEF_CMD (S_ANY, 0, "PRESERVE", N_("Preserve"), cmd_preserve)
DEF_CMD (S_ANY, F_ABBREV, "Q", N_("Quit"), cmd_finish)
DEF_CMD (S_ANY, 0, "QUIT", N_("Quit"), cmd_finish)
DEF_CMD (S_ANY, 0, "RESTORE", N_("Restore"), cmd_restore)
DEF_CMD (S_ANY, 0, "SET", N_("Set"), cmd_set)
DEF_CMD (S_ANY, 0, "SHOW", N_("Show"), cmd_show)
DEF_CMD (S_ANY, 0, "SUBTITLE", N_("Subtitle"), cmd_subtitle)
DEF_CMD (S_ANY, 0, "SYSFILE INFO", N_("Sysfile Info"), cmd_sysfile_info)
DEF_CMD (S_ANY, 0, "TITLE", N_("Title"), cmd_title)

/* Commands that define (or replace) the active dataset. */
DEF_CMD (S_INITIAL | S_DATA, 0, "ADD FILES", N_("Add Files"), cmd_add_files)
DEF_CMD (S_INITIAL | S_DATA | S_INPUT_PROGRAM | S_FILE_TYPE | S_NESTED_INPUT_PROGRAM,
         0, "DATA LIST", N_("Data List"), cmd_data_list)
DEF_CMD (S_INITIAL | S_DATA, 0, "GET", N_("Get"), cmd_get)
DEF_CMD (S_INITIAL | S_DATA, 0, "GET DATA", N_("Get Data"), cmd_get_data)
DEF_CMD (S_INITIAL | S_DATA, 0, "IMPORT", N_("Import"), cmd_import)
DEF_CMD (S_INITIAL | S_DATA, 0, "INPUT PROGRAM", N_("Input Program"), cmd_input_program)
DEF_CMD (S_INITIAL | S_DATA, 0, "MATCH FILES", N_("Match Files"), cmd_match_files)
DEF_CMD (S_INITIAL | S_DATA, 0, "MATRIX DATA", N_("Matrix Data"), cmd_matrix_data)
DEF_CMD (S_INITIAL | S_DATA, 0, "UPDATE", N_("Update"), cmd_update)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET ACTIVATE", N_("Dataset Activate"), cmd_dataset_activate)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET DECLARE", N_("Dataset Declare"), cmd_dataset_declare)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET CLOSE", N_("Dataset Close"), cmd_dataset_close)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET COPY", N_("Dataset Copy"), cmd_dataset_copy)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET NAME", N_("Dataset Name"), cmd_dataset_name)
DEF_CMD (S_INITIAL | S_DATA, 0, "DATASET DISPLAY", N_("Dataset Display"), cmd_dataset_display)

/* Utilities that may appear after active file definition or within INPUT PROGRAM. */
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "ADD VALUE LABELS", N_("Add Value Labels"), cmd_add_value_labels)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "ADD DOCUMENT", N_("Add Document"), cmd_add_documents)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "APPLY DICTIONARY", N_("Apply Dictionary"), cmd_apply_dictionary)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "DATAFILE ATTRIBUTE", N_("Datafile Attribute"), cmd_datafile_attribute)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "DISPLAY", N_("Display"), cmd_display)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, F_TESTING, "DISPLAY VARIABLE SETS", N_("Display Variable Sets"), cmd_display_variable_sets)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "DOCUMENT", N_("Document"), cmd_document)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "DROP DOCUMENTS", N_("Drop Documents"), cmd_drop_documents)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "FORMATS", N_("Formats"), cmd_formats)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "LEAVE", N_("Leave"), cmd_leave)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "MISSING VALUES", N_("Missing Values"), cmd_missing_values)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "MRSETS", N_("Multiple Response Sets"), cmd_mrsets)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "PRINT FORMATS", N_("Print Formats"), cmd_print_formats)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "SPLIT FILE", N_("Split File"), cmd_split_file)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VALUE LABELS", N_("Value Labels"), cmd_value_labels)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE ALIGNMENT", N_("Variable Alignment"), cmd_variable_alignment)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE ATTRIBUTE", N_("Variable Attribute"), cmd_variable_attribute)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE LABELS", N_("Variable Labels"), cmd_variable_labels)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE LEVEL", N_("Variable Level"), cmd_variable_level)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE ROLE", N_("Variable Role"), cmd_variable_role)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VARIABLE WIDTH", N_("Variable Width"), cmd_variable_width)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "VECTOR", N_("Vector"), cmd_vector)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "WEIGHT", N_("Weight"), cmd_weight)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "WRITE FORMATS", N_("Write Formats"), cmd_write_formats)

/* Transformations. */
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "BREAK", N_("Break"), cmd_break)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "COMPUTE", N_("Compute"), cmd_compute)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "DO IF", N_("Do If"), cmd_do_if)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "IF", N_("If"), cmd_if)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "LOOP", N_("Loop"), cmd_loop)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "NUMERIC", N_("Numeric"), cmd_numeric)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "PRINT EJECT", N_("Print Eject"), cmd_print_eject)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "PRINT SPACE", N_("Print Space"), cmd_print_space)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "PRINT", N_("Print"), cmd_print)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "RECODE", N_("Recode"), cmd_recode)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "STRING", N_("String"), cmd_string)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "WRITE", N_("Write"), cmd_write)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, F_ENHANCED, "XEXPORT", N_("XExport"), cmd_xexport)
DEF_CMD (S_DATA | S_INPUT_PROGRAM | S_NESTED_ANY, 0, "XSAVE", N_("XSave"), cmd_xsave)

/* restricted Transformations. */
DEF_CMD (S_DATA | S_NESTED_DATA, 0, "SELECT IF", N_("Select If"), cmd_select_if)

/* Commands that may appear after active dataset definition. */
DEF_CMD (S_DATA, 0, "AGGREGATE", N_("Aggregate"), cmd_aggregate)
DEF_CMD (S_DATA, 0, "AUTORECODE", N_("Autorecode"), cmd_autorecode)
DEF_CMD (S_DATA, 0, "BEGIN DATA", N_("Begin Data"), cmd_begin_data)
DEF_CMD (S_DATA, 0, "COUNT", N_("Count"), cmd_count)
DEF_CMD (S_DATA, 0, "CROSSTABS", N_("Crosstabs"), cmd_crosstabs)
DEF_CMD (S_DATA, 0, "CTABLES", N_("Ctables"), cmd_ctables)
DEF_CMD (S_DATA, 0, "CORRELATIONS", N_("Correlations"), cmd_correlations)
DEF_CMD (S_DATA, 0, "DELETE VARIABLES", N_("Delete Variables"), cmd_delete_variables)
DEF_CMD (S_DATA, 0, "DESCRIPTIVES", N_("Descriptives"), cmd_descriptives)
DEF_CMD (S_DATA, 0, "EXAMINE", N_("Examine"), cmd_examine)
DEF_CMD (S_DATA, 0, "EXECUTE", N_("Execute"), cmd_execute)
DEF_CMD (S_DATA, 0, "EXPORT", N_("Export"), cmd_export)
DEF_CMD (S_DATA, 0, "FACTOR", N_("Factor"), cmd_factor)
DEF_CMD (S_DATA, 0, "FILTER", N_("Filter"), cmd_filter)
DEF_CMD (S_DATA, 0, "FLIP", N_("Flip"), cmd_flip)
DEF_CMD (S_DATA, 0, "FREQUENCIES", N_("Frequencies"), cmd_frequencies)
DEF_CMD (S_DATA, 0, "GLM", N_("GLM"), cmd_glm)
DEF_CMD (S_DATA, 0, "GRAPH", N_("Graph"), cmd_graph)
DEF_CMD (S_DATA, 0, "LIST", N_("List"), cmd_list)
DEF_CMD (S_DATA, 0, "LOGISTIC REGRESSION", N_("Logistic Regression"), cmd_logistic)
DEF_CMD (S_DATA, 0, "MEANS", N_("Means"), cmd_means)
DEF_CMD (S_DATA, 0, "NPAR TESTS", N_("Nonparametric Tests"), cmd_npar_tests)
DEF_CMD (S_DATA, 0, "ONEWAY", N_("Oneway"), cmd_oneway)
DEF_CMD (S_DATA, 0, "PEARSON CORRELATIONS", N_("Pearson Correlations"), cmd_correlations)
DEF_CMD (S_DATA, 0, "QUICK CLUSTER", N_("Quick Cluster"), cmd_quick_cluster)
DEF_CMD (S_DATA, 0, "RANK", N_("Rank"), cmd_rank)
DEF_CMD (S_DATA, 0, "REGRESSION", N_("Regression"),  cmd_regression)
DEF_CMD (S_DATA, 0, "RELIABILITY", N_("Reliability"), cmd_reliability)
DEF_CMD (S_DATA, 0, "RENAME VARIABLES", N_("Rename Variables"), cmd_rename_variables)
DEF_CMD (S_DATA, 0, "ROC", N_("ROC"), cmd_roc)
DEF_CMD (S_DATA, 0, "SAMPLE", N_("Sample"), cmd_sample)
DEF_CMD (S_DATA, 0, "SAVE", N_("Save"), cmd_save)
DEF_CMD (S_DATA, 0, "SAVE DATA COLLECTION", N_("Save Data Collection"), cmd_save_data_collection)
DEF_CMD (S_DATA, 0, "SAVE TRANSLATE", N_("Save Translate"), cmd_save_translate)
DEF_CMD (S_DATA, 0, "SORT CASES", N_("Sort Cases"), cmd_sort_cases)
DEF_CMD (S_DATA, 0, "SORT VARIABLES", N_("Sort Variables"), cmd_sort_variables)
DEF_CMD (S_DATA, 0, "T-TEST", N_("T-Test"), cmd_t_test)
DEF_CMD (S_DATA, 0, "TEMPORARY", N_("Temporary"), cmd_temporary)
DEF_CMD (S_DATA, 0, "USE", N_("Use"), cmd_use)

/* Commands valid only inside INPUT PROGRAM. */
DEF_CMD (S_INPUT_PROGRAM | S_NESTED_INPUT_PROGRAM, 0, "END CASE", N_("End Case"), cmd_end_case)
DEF_CMD (S_INPUT_PROGRAM | S_NESTED_INPUT_PROGRAM, 0, "END FILE", N_("End File"), cmd_end_file)
DEF_CMD (S_INPUT_PROGRAM | S_NESTED_INPUT_PROGRAM, 0, "REREAD", N_("Reread"), cmd_reread)

/* Not really commands, but part of LOOP and DO IF syntax. */
DEF_CMD (S_ANY, 0, "ELSE IF", N_("Else If"), cmd_inside_do_if)
DEF_CMD (S_ANY, 0, "ELSE", N_("Else"), cmd_inside_do_if)
DEF_CMD (S_ANY, 0, "END IF", N_("End If"), cmd_inside_do_if)
DEF_CMD (S_ANY, 0, "END LOOP", N_("End Loop"), cmd_inside_loop)

/* Commands for testing PSPP. */
DEF_CMD (S_ANY, F_TESTING, "DEBUG EXPAND", "DEBUG EXPAND", cmd_debug_expand)
DEF_CMD (S_ANY, F_TESTING, "DEBUG EVALUATE", "DEBUG EVALUATE", cmd_debug_evaluate)
DEF_CMD (S_ANY, F_TESTING, "DEBUG FORMAT GUESSER", "DEBUG FORMAT GUESSER", cmd_debug_format_guesser)
DEF_CMD (S_ANY, F_TESTING, "DEBUG MOMENTS", "DEBUG MOMENTS", cmd_debug_moments)
DEF_CMD (S_ANY, F_TESTING, "DEBUG PAPER SIZE", "DEBUG PAPER SIZE", cmd_debug_paper_size)
DEF_CMD (S_ANY, F_TESTING, "DEBUG POOL", "DEBUG POOL", cmd_debug_pool)
DEF_CMD (S_ANY, F_TESTING, "DEBUG FLOAT FORMAT", "DEBUG FLOAT FORMAT", cmd_debug_float_format)
DEF_CMD (S_ANY, F_TESTING, "DEBUG XFORM FAIL", "DEBUG XFORM FAIL", cmd_debug_xform_fail)
DEF_CMD (S_DATA, F_TESTING, "DEBUG MATRIX READ", "DEBUG MATRIX READ", cmd_debug_matrix_read)

/* Unimplemented commands. */
UNIMPL_CMD ("2SLS", "Two stage least squares regression")
UNIMPL_CMD ("ACF", "Autocorrelation function")
UNIMPL_CMD ("ALSCAL", "Multidimensional scaling")
UNIMPL_CMD ("ANACOR", "Correspondence analysis")
UNIMPL_CMD ("ANOVA", "Factorial analysis of variance")
UNIMPL_CMD ("CASEPLOT", "Plot time series")
UNIMPL_CMD ("CASESTOVARS", "Restructure complex data")
UNIMPL_CMD ("CATPCA", "Categorical principle components analysis")
UNIMPL_CMD ("CATREG", "Categorical regression")
UNIMPL_CMD ("CCF", "Time series cross correlation")
UNIMPL_CMD ("CLEAR TRANSFORMATIONS", "Clears transformations from active dataset")
UNIMPL_CMD ("CLUSTER", "Hierarchical clustering")
UNIMPL_CMD ("CONJOINT", "Analyse full concept data")
UNIMPL_CMD ("CORRESPONDENCE", "Show correspondence")
UNIMPL_CMD ("COXREG", "Cox proportional hazards regression")
UNIMPL_CMD ("CREATE", "Create time series data")
UNIMPL_CMD ("CSDESCRIPTIVES", "Complex samples descriptives")
UNIMPL_CMD ("CSGLM", "Complex samples GLM")
UNIMPL_CMD ("CSLOGISTIC", "Complex samples logistic regression")
UNIMPL_CMD ("CSPLAN", "Complex samples design")
UNIMPL_CMD ("CSSELECT", "Select complex samples")
UNIMPL_CMD ("CSTABULATE", "Tabulate complex samples")
UNIMPL_CMD ("CURVEFIT", "Fit curve to line plot")
UNIMPL_CMD ("DATE", "Create time series data")
UNIMPL_CMD ("DETECTANOMALY", "Find unusual cases")
UNIMPL_CMD ("DISCRIMINANT", "Linear discriminant analysis")
UNIMPL_CMD ("EDIT", "obsolete")
UNIMPL_CMD ("END FILE TYPE", "Ends complex data input")
UNIMPL_CMD ("FILE TYPE", "Complex data input")
UNIMPL_CMD ("FIT", "Goodness of Fit")
UNIMPL_CMD ("GENLOG", "Categorical model fitting")
UNIMPL_CMD ("GET TRANSLATE", "Read other file formats")
UNIMPL_CMD ("GGRAPH", "Custom defined graphs")
UNIMPL_CMD ("HILOGLINEAR", "Hierarchical loglinear models")
UNIMPL_CMD ("HOMALS", "Homogeneity analysis")
UNIMPL_CMD ("IGRAPH", "Interactive graphs")
UNIMPL_CMD ("INFO", "Local Documentation")
UNIMPL_CMD ("KEYED DATA LIST", "Read nonsequential data")
UNIMPL_CMD ("KM", "Kaplan-Meier")
UNIMPL_CMD ("LOGLINEAR", "General model fitting")
UNIMPL_CMD ("MANOVA", "Multivariate analysis of variance")
UNIMPL_CMD ("MAPS", "Geographical display")
UNIMPL_CMD ("MIXED", "Mixed linear models")
UNIMPL_CMD ("MODEL CLOSE", "Close server connection")
UNIMPL_CMD ("MODEL HANDLE", "Define server connection")
UNIMPL_CMD ("MODEL LIST", "Show existing models")
UNIMPL_CMD ("MODEL NAME", "Specify model label")
UNIMPL_CMD ("MULTIPLE CORRESPONDENCE", "Multiple correspondence analysis")
UNIMPL_CMD ("MULT RESPONSE", "Multiple response analysis")
UNIMPL_CMD ("MVA", "Missing value analysis")
UNIMPL_CMD ("NAIVEBAYES", "Small sample bayesian prediction")
UNIMPL_CMD ("NLR", "Non Linear Regression")
UNIMPL_CMD ("NOMREG", "Multinomial logistic regression")
UNIMPL_CMD ("NONPAR CORR", "Nonparametric correlation")
UNIMPL_CMD ("NUMBERED", "")
UNIMPL_CMD ("OLAP CUBES", "On-line analytical processing")
UNIMPL_CMD ("OMS", "Output management")
UNIMPL_CMD ("ORTHOPLAN", "Orthogonal effects design")
UNIMPL_CMD ("OVERALS", "Nonlinear canonical correlation")
UNIMPL_CMD ("PACF", "Partial autocorrelation")
UNIMPL_CMD ("PARTIAL CORR", "Partial correlation")
UNIMPL_CMD ("PLANCARDS", "Conjoint analysis planning")
UNIMPL_CMD ("PLUM", "Estimate ordinal regression models")
UNIMPL_CMD ("POINT", "Marker in keyed file")
UNIMPL_CMD ("PPLOT", "Plot time series variables")
UNIMPL_CMD ("PREDICT", "Specify forecast period")
UNIMPL_CMD ("PREFSCAL", "Multidimensional unfolding")
UNIMPL_CMD ("PRINCALS", "PCA by alternating least squares")
UNIMPL_CMD ("PROBIT", "Probit analysis")
UNIMPL_CMD ("PROCEDURE OUTPUT", "Specify output file")
UNIMPL_CMD ("PROXIMITIES", "Pairwise similarity")
UNIMPL_CMD ("PROXSCAL", "Multidimensional scaling of proximity data")
UNIMPL_CMD ("RATIO STATISTICS", "Descriptives of ratios")
UNIMPL_CMD ("READ MODEL", "Read new model")
UNIMPL_CMD ("RECORD TYPE", "Defines a type of record within FILE TYPE")
UNIMPL_CMD ("REFORMAT", "Read obsolete files")
UNIMPL_CMD ("REPEATING DATA", "Specify multiple cases per input record")
UNIMPL_CMD ("REPORT", "Pretty print working file")
UNIMPL_CMD ("RMV", "Replace missing values")
UNIMPL_CMD ("SCRIPT", "Run script file")
UNIMPL_CMD ("SEASON", "Estimate seasonal factors")
UNIMPL_CMD ("SELECTPRED", "Select predictor variables")
UNIMPL_CMD ("SPCHART", "Plot control charts")
UNIMPL_CMD ("SPECTRA", "Plot spectral density")
UNIMPL_CMD ("STEMLEAF", "Plot stem-and-leaf display")
UNIMPL_CMD ("SUMMARIZE", "Univariate statistics")
UNIMPL_CMD ("SURVIVAL", "Survival analysis")
UNIMPL_CMD ("TDISPLAY", "Display active models")
UNIMPL_CMD ("TREE", "Create classification tree")
UNIMPL_CMD ("TSAPPLY", "Apply time series model")
UNIMPL_CMD ("TSET", "Set time sequence variables")
UNIMPL_CMD ("TSHOW", "Show time sequence variables")
UNIMPL_CMD ("TSMODEL", "Estimate time series model")
UNIMPL_CMD ("TSPLOT", "Plot time sequence variables")
UNIMPL_CMD ("TWOSTEP CLUSTER", "Cluster observations")
UNIMPL_CMD ("UNIANOVA", "Univariate analysis")
UNIMPL_CMD ("UNNUMBERED", "obsolete")
UNIMPL_CMD ("VALIDATEDATA", "Identify suspicious cases")
UNIMPL_CMD ("VARCOMP", "Estimate variance")
UNIMPL_CMD ("VARSTOCASES", "Restructure complex data")
UNIMPL_CMD ("VERIFY", "Report time series")
UNIMPL_CMD ("WLS", "Weighted least squares regression")
UNIMPL_CMD ("XGRAPH", "High resolution charts")
